#include "block.h"
#include <QJsonArray>
#include <layout.h>

Block::Block(uint32_t id):
	id_(id)
{

}

void Block::addBorder(std::shared_ptr<BlockBorder> border)
{
	borders_.push_back(border);
	std::shared_ptr<Block> block = border->getOtherBlock(this);
	if(!block || block.get() == this)
	{
		qWarning()<<__func__<<"no other block in border"<<border->id();
		return;
	}
	connect(block.get(), &Block::blockedChanged, this, &Block::checkWaits);
}

bool Block::blocked()
{
	if(trains_.empty() && waits_.empty())
		return false;
	return true;
}

bool Block::ownsTrain(std::weak_ptr<Train> train)
{
	std::shared_ptr<Train> trainPtr = train.lock();
	if(!trainPtr)
		return false;

	for(auto ownedTrain : trains_)
	{
		if(ownedTrain.lock() == trainPtr)
			return true;
	}
	return false;
}

bool Block::ownsBorder(std::shared_ptr<BlockBorder> border)
{
	if(!border)
		return false;
	for(std::shared_ptr<BlockBorder> borderTest : borders_)
	{
		if(borderTest == border)
			return true;
	}
	return false;
}

void Block::checkWaits(bool ign)
{
	qDebug()<<__func__;
	(void)ign;
	bool wasBlocked = blocked();
	for(std::vector<TrainWait>::iterator iter = waits_.begin(); iter != waits_.end();)
	{
		std::shared_ptr<Block> block = iter->targetBlock.lock();
		std::shared_ptr<Train> train = iter->train.lock();
		qDebug()<<__func__<<"trying to push train"<<train->getTrainId()<<"to block"<<block->id();
		if(block && train && train->suspendend() && block->pushTrain(iter->train))
		{
			train->resume();
			std::shared_ptr<BlockBorder> border = iter->border.lock();
			if(border)
				border->updateSignals();
			iter = waits_.erase(iter);
		}
		else if(!block || !train || !train->suspendend())
		{
			iter = waits_.erase(iter);
		}
		else
		{
			++iter;
		}
	}
	if(wasBlocked != blocked())
		blockedChanged(!wasBlocked);
}

void Block::updateBorders()
{
	for(std::shared_ptr<BlockBorder> broder : borders_)
	{
		broder->updateSignals();
	}
}

bool Block::pushTrain(std::weak_ptr<Train> train)
{
	if(blocked())
		return false;

	addTrain(train);
	return true;
}

void Block::addTrain(std::weak_ptr<Train> train)
{
	std::shared_ptr<Train> trainPtr = train.lock();
	if(!trainPtr)
		return;
	qDebug()<<"Train"<<trainPtr->getTrainId()<<"added to block"<<id_;
	bool wasBlocked = blocked();
	trains_.push_back(train);
	if(wasBlocked != blocked())
		blockedChanged(!wasBlocked);
	updateBorders();
}

void Block::unsuspendedTrain(uint32_t id, int direction)
{
	qDebug()<<"Train with item id"<<id<<"reports unsuspended";
	bool wasBlocked = blocked();
	for(std::vector<TrainWait>::iterator iter = waits_.begin(); iter != waits_.end(); ++iter)
	{
		std::shared_ptr<Train> workTrain = iter->train.lock();
		if(!workTrain)
		{
			waits_.erase(iter);
			unsuspendedTrain(id, direction);
			break;
		}
		if(workTrain->id() == id)
		{
			std::shared_ptr<Block> block = iter->targetBlock.lock();
			disconnect(workTrain.get(), &Train::unsuspended, this, &Block::unsuspendedTrain);
			if(iter->direction == direction)
			{
				if(block)
					block->addTrain(workTrain);
			}
			else
			{
				addTrain(workTrain);
			}
			waits_.erase(iter);
			break;
		}
	}
	if(wasBlocked != blocked())
		blockedChanged(!wasBlocked);
	updateBorders();
}

void Block::removeTrain(std::weak_ptr<Train> train)
{
	std::shared_ptr<Train> trainPtr = train.lock();
	if(!trainPtr)
		return;

	for(std::vector<std::weak_ptr<Train>>::iterator iter = trains_.begin(); iter != trains_.end(); ++iter)
	{
		if(iter->lock() == trainPtr)
		{
			qDebug()<<"Train"<<trainPtr->getTrainId()<<"removed from to block"<<id_;
			trains_.erase(iter);
			break;
		}
	}
}

void Block::trainArrivedAtBorder(std::weak_ptr<BlockBorder> borderIn, std::weak_ptr<Train> trainIn)
{
	qDebug()<<__func__<<"block"<<id();
	bool wasBlocked = blocked();
	std::shared_ptr<BlockBorder> border = borderIn.lock();
	if(!border)
	{
		qWarning()<<"border has expired in"<<__func__;
		return;
	}

	if(!ownsBorder(border))
		return;

	qDebug()<<"Block"<<id_<<"owns subject border";

	if(!ownsTrain(trainIn))
		return;

	qDebug()<<"Block"<<id_<<"owns subject train";

	std::shared_ptr block = border->getOtherBlock(this);
	if(block)
	{
		std::shared_ptr<Train> workTrain = trainIn.lock();
		if(workTrain)
		{
			if(!block->pushTrain(trainIn))
			{
				workTrain->suspend();
				TrainWait wait;
				wait.train = trainIn;
				wait.direction = workTrain->getDirection();
				wait.targetBlock = block;
				wait.border = border;
				connect(workTrain.get(), &Train::unsuspended, this, &Block::unsuspendedTrain);
				waits_.push_back(wait);
				qDebug()<<"Train"<<workTrain->getTrainId()<<"is wating at border for block"<<block->id();
				if(!block->trains_.empty() && block->trains_[0].lock())
					qDebug()<<"for train"<<block->trains_[0].lock()->getTrainId()<<"to leave";
			}
			removeTrain(workTrain);
		}
	}
	if(wasBlocked != blocked())
		blockedChanged(blocked());
	updateBorders();
}

void Block::store(QJsonObject &json)
{
	json["Id"] = static_cast<double>(id_);
	QJsonArray borders;
	for(auto& border : borders_)
	{
		QJsonObject borderObject;
		borderObject["BorderId"] =  static_cast<double>(border->id());
		borders.append(borderObject);
	}
	json["Borders"] = borders;
}

void Block::load(const QJsonObject &json)
{
	id_ = static_cast<uint32_t>(json["Id"].toDouble(0));
}

void Block::populate(const QJsonObject& json, Layout* layout)
{
	borders_.clear();
	const QJsonArray bordersArray(json["Borders"].toArray(QJsonArray()));
	for(const auto& arrayObj : bordersArray)
	{
		if(arrayObj.isObject())
		{
			uint32_t borderId = static_cast<uint32_t>(arrayObj.toObject()["BorderId"].toDouble(INT32_MAX));
			std::shared_ptr<BlockBorder> border = layout->getBorder(borderId);
			if(border)
				addBorder(border);
			else if(borderId == INT32_MAX)
				qWarning()<<"BorderId field is missing in border array for block"<<id_;
			else
				qWarning()<<"border"<<borderId<<"dosent exist but is needed by block"<<id_;
		}
	}
}
