#include "layout.h"
#include <memory>
#include <QJsonArray>

Layout::Layout(QObject *parent): QObject{parent}
{
}

void Layout::itemAdded(std::weak_ptr<Item> itemIn)
{
	std::shared_ptr<Item> item = itemIn.lock();
	if(!item)
		return;
	std::shared_ptr<Train> train = std::dynamic_pointer_cast<Train>(item);
	std::shared_ptr<Signal> signal = std::dynamic_pointer_cast<Signal>(item);
}

void Layout::removeTrainFromAllBlocks(std::shared_ptr<Train> train)
{
	for(std::shared_ptr<Block> block : blocks_)
		block->removeTrain(train);
}

void Layout::registerTrainInLimbo(std::shared_ptr<Train> train, std::shared_ptr<BlockBorder> border)
{
	for(auto iter = trainLimbo_.begin(); iter != trainLimbo_.end(); ++iter)
	{
		std::shared_ptr<Train> limboTrain = iter->first.lock();
		std::shared_ptr<BlockBorder> limboBorder = iter->second.lock();
		if(!limboTrain || !limboBorder)
		{
			trainLimbo_.erase(iter);
			registerTrainInLimbo(train, border);
			return;
		}

		if(*train == *limboTrain)
		{
			if(border == limboBorder)
				return;

			std::shared_ptr<Block> overlap = border->getOverlap(limboBorder.get());
			if(overlap)
			{
				qDebug()<<"Train"<<train->getTrainId()<<"removed from limbo and added to block"<<overlap->id()<<"while crossing border"<<border->id();
				overlap->addTrain(train);
				overlap->trainArrivedAtBorder(border, train);
				trainLimbo_.erase(iter);
				return;
			}
			else
			{
				iter->second = border;
				return;
			}

		}
	}
	std::shared_ptr<Block> first = border->getBlocks().first.lock();
	std::shared_ptr<Block> second = border->getBlocks().second.lock();

	qDebug()<<"Train"<<train->getTrainId()<<"added to limbo between block"<<
			  (first ? QString::number(first->id()) : "invalid")<<"and block"<<(second ? QString::number(second->id()) : "invalid");
	trainLimbo_.push_back(std::pair<std::weak_ptr<Train>, std::weak_ptr<BlockBorder>>(train, border));
}

void Layout::trainArrivedAtReader(uint8_t reader, std::shared_ptr<Train> train, int tagType)
{
	std::shared_ptr<BlockBorder> border;
	for(std::shared_ptr<BlockBorder> borderTest : borders_)
	{
		if(borderTest->getReader() == reader)
		{
			border = borderTest;
			break;
		}
	}
	if(!border)
	{
		qWarning()<<"reader "<<reader<<"is not registerd with any border";
		return;
	}
	qDebug()<<"Train"<<train->getTrainId()<<"arrived at border"<<border->id();

	bool trainHandled = false;
	for(std::shared_ptr<Block> block : blocks_)
	{
		if(block->ownsTrain(train))
		{
			trainHandled = true;
			if(block->ownsBorder(border))
			{
				block->trainArrivedAtBorder(border, train);
			}
			else
			{
				removeTrainFromAllBlocks(train);
				registerTrainInLimbo(train, border);
			}
			break;
		}
	}

	if(!trainHandled)
		registerTrainInLimbo(train, border);
}

void Layout::addBlock(std::shared_ptr<Block> block)
{
	blocks_.push_back(block);
}

void Layout::addBorder(std::shared_ptr<BlockBorder> border)
{
	borders_.push_back(border);
}

std::shared_ptr<BlockBorder> Layout::getBorder(uint32_t id)
{
	for(const std::shared_ptr<BlockBorder>& border : borders_)
	{
		if(border->id() == id)
			return border;
	}
	return std::shared_ptr<BlockBorder>();
}

std::shared_ptr<Block> Layout::getBlock(uint32_t id)
{
	for(const std::shared_ptr<Block>& block : blocks_)
	{
		if(block->id() == id)
			return block;
	}
	return std::shared_ptr<Block>();
}

void Layout::store(QJsonObject& json)
{
	QJsonArray blockArray;
	for(const std::shared_ptr<Block>& block : blocks_)
	{
		QJsonObject object;
		block->store(object);
		blockArray.append(object);
	}
	json["Blocks"] = blockArray;

	QJsonArray borderArray;
	for(const std::shared_ptr<BlockBorder>& border : borders_)
	{
		QJsonObject object;
		border->store(object);
		borderArray.append(object);
	}
	json["Borders"] = borderArray;
}

void Layout::load(const QJsonObject& json)
{
	std::vector<QJsonObject> blockJsonObjects;
	const QJsonArray blockArray(json["Blocks"].toArray(QJsonArray()));
	for(const auto& object : blockArray)
	{
		if(object.isObject())
		{
			const QJsonObject jsonObject = object.toObject();
			blockJsonObjects.push_back(jsonObject);
			std::shared_ptr<Block> block(new Block);
			block->load(jsonObject);
			blocks_.push_back(block);
		}
	}

	std::vector<QJsonObject> borderJsonObjects;
	const QJsonArray borderArray(json["Borders"].toArray(QJsonArray()));
	for(const auto& object : borderArray)
	{
		if(object.isObject())
		{
			const QJsonObject jsonObject = object.toObject();
			borderJsonObjects.push_back(jsonObject);
			std::shared_ptr<BlockBorder> border(new BlockBorder);
			border->load(object.toObject());
			borders_.push_back(border);
		}
	}

	for(size_t i = 0; i < blocks_.size(); ++i)
	{
		blocks_[i]->populate(blockJsonObjects[i], this);
	}

	for(std::shared_ptr<BlockBorder>& border : borders_)
	{
		std::shared_ptr<Block> first;
		std::shared_ptr<Block> second;
		for(std::shared_ptr<Block>& block : blocks_)
		{
			if(block->ownsBorder(border))
			{
				if(!first)
					first = block;
				else if(!second)
					second = block;
				else if(!second)
					qWarning()<<"border with id"<<border->id()<<"is assigned to more than 2 blocks";
			}
		}
		if(!first || !second)
			qWarning()<<"border with id"<<border->id()<<"is assigned to less than than 2 blocks";
		border->setBlocks(std::pair<std::weak_ptr<Block>, std::weak_ptr<Block>>(first, second));
	}

	for(size_t i = 0; i < blocks_.size(); ++i)
	{
		blocks_[i]->populate(blockJsonObjects[i], this);
		if(blocks_[i]->getBorders().empty())
			qWarning()<<"block with id"<<blocks_[i]->id()<<"doset have at least one border";
	}
}
